;display controller for Sharp M078CKA-A3QKLA0057 (240x64 pixels)
;ATTiny2313 @ 7.3728 MHz (crystal), lfuse=0xDD, hfuse=0xDB

;PORTS
;  PB0-7 = Display Data Bit 0-7
;  PD0   = RS232 Rx Input
;  PD1   = unused
;  PD2   = Display Reset (active low)
;  PD3   = Display Select (active low)
;  PD4   = Display RS (high for command)
;  PD5   = Display E (active high)
;  PD6   = unused

.include "../include/tn2313def.inc"

.equ	LCD_DATA_PORT = PORTB
.equ	LCD_CTRL_PORT = PORTD
.equ	LCD_E = 5
.equ	LCD_RS = 4
.equ	LCD_CS = 3
.equ	LCD_R = 2

;R1:R0 = scratch registers (can be modified in any function except interrupt)

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	lcd_x = R19
.def	lcd_y = R20
.def	mask = R21
.def	byte = R22
.def	textlen = R23
.def	bufstart = R24
.def	bufend = R25
.def	buflen = R26

;R29:R28 = Y pointer (RAM)
;R31:R30 = Z pointer (Flash)

;===============================================================================

.dseg
	RAM_Text: .byte 40 ;one line of text
	RAM_Buffer: .byte 64 ;UART Rx buffer

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org URXCaddr
	rjmp	uart_rx
	
charset:
	.include "charset-cp437.asm"
	
;===============================================================================

uart_rx:
	;incoming character interrupt: add char to ring buffer
	push	YL
	push	temp
	in	temp, SREG
	push	temp
	
	in	temp, UDR
	cpi	buflen, 64 ;buffer full: discard char
	brsh	uart_rx_end
	ldi	YL, RAM_Buffer
	add	YL, bufend
	st	Y, temp
	inc	bufend
	inc	buflen
	andi	bufend, 0x3F
	
uart_rx_end:
	pop	temp
	out	SREG, temp
	pop	temp
	pop	YL
	reti

;===============================================================================

reset:
	;set stackpointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;init ports
	ldi	temp, 0xFF
	out	DDRB, temp
	ldi	temp, 0x7E
	out	DDRD, temp
	ldi	temp, 1<<LCD_R | 1<<LCD_CS | 0x01 ;enable pullup for UART Rx
	out	PORTD, temp
	
	;init UART
	ldi	temp, 1<<RXEN | 1<<RXCIE
	out	UCSRB, temp
;	ldi	temp, 47 ;9600 Baud @ 7.3728 MHz
	ldi	temp, 11 ;38400 Baud @ 7.3728 MHz
	out	UBRRL, temp
	
	;init display
	rcall	lcd_wait
	cbi	LCD_CTRL_PORT, LCD_R ;reset
	rcall	lcd_wait
	sbi	LCD_CTRL_PORT, LCD_R
	rcall	lcd_wait
	cbi	LCD_CTRL_PORT, LCD_CS
	ldi	byte, 0xF0
	rcall	lcd_cmd
	ldi	byte, 0x40
	rcall	lcd_cmd
	ldi	byte, 0x50
	rcall	lcd_cmd
	ldi	byte, 0x61
	rcall	lcd_cmd
	ldi	byte, 0x70
	rcall	lcd_cmd
	ldi	byte, 0xF0
	rcall	lcd_cmd
	ldi	byte, 0x81
	rcall	lcd_cmd
	ldi	byte, 0x92
	rcall	lcd_cmd
	ldi	byte, 0xA1
	rcall	lcd_cmd
	ldi	byte, 0xB2
	rcall	lcd_cmd
	ldi	byte, 0xDE
	rcall	lcd_cmd
	ldi	byte, 0xE0
	rcall	lcd_cmd
	ldi	byte, 0xF1
	rcall	lcd_cmd
	ldi	byte, 0xE0
	rcall	lcd_cmd
	ldi	byte, 0xF0
	rcall	lcd_cmd
	rcall	lcd_clear
	
	;init registers
	ldi	textlen, 0
	ldi	lcd_y, 0
	ldi	bufstart, 0
	ldi	bufend, 0
	ldi	buflen, 0
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------

loop:
	;get char from UART buffer
	cpi	buflen, 0
	breq	loop
	ldi	YL, RAM_Buffer
	add	YL, bufstart
	ld	temp, Y
	dec	buflen
	inc	bufstart
	andi	bufstart, 0x3F
	
	cpi	temp, 0x20
	brsh	incoming_char
	cpi	temp, 0x02
	breq	incoming_home
	cpi	temp, 0x08
	breq	incoming_backspace
	cpi	temp, 0x0a
	breq	incoming_lf
	cpi	temp, 0x0c
	breq	incoming_ff
	;ignore other control chars
	rjmp	loop
	
incoming_home:
	;Start of Text: return cursor home
	;NOTE: this is not the usual behaviour of ASCII code 0x02!
	ldi	lcd_y, 0
	ldi	textlen, 0
	rjmp	loop

incoming_backspace:
	;Backspace: remove last char from text (if text length > 0)
	cpi	textlen, 0
	breq	loop
	dec	textlen
	rjmp	loop

incoming_lf:
	;Line Feed: fill line with spaces if < 40 chars and display it
	cpi	lcd_y, 64 ;LCD already full?
	brsh	loop
	ldi	YL, RAM_Text
	add	YL, textlen
	ldi	temp, ' '
incoming_lf_fill:
	cpi	YL, RAM_Text+40
	brsh	incoming_lf_send
	st	Y+, temp
	rjmp	incoming_lf_fill
incoming_lf_send:
	rcall	lcd_text
	ldi	textlen, 0
	rjmp	loop
	
incoming_ff:
	;Form Feed: clear display
	rcall	lcd_clear
	ldi	lcd_y, 0
	ldi	textlen, 0
	rjmp	loop
	
incoming_char:
	;received char from UART: store to buffer RAM
	cpi	textlen, 40 ;end of line?
	brsh	loop
	ldi	YL, RAM_Text
	add	YL, textlen
	st	Y, temp
	inc	textlen
	rjmp	loop

;===============================================================================

lcd_text:
	;display one line of text from RAM
	ldi	YL, RAM_Text
	rcall	lcd_pos
	mov	temp, lcd_y ;calculate bit mask for charset bytes
	andi	temp, 0x07
	ldi	mask, 0x01
	cpi	temp, 0
	breq	lcd_text_mask_end
lcd_text_mask:
	lsl	mask
	dec	temp
	brne	lcd_text_mask
lcd_text_mask_end:
	
	ldi	lcd_x, 40*6 ;40 chars * 6 pixels each
lcd_text_line_loop:
	ld	ZL, Y+
	subi	ZL, 0x20 ;charset starts with char 0x20
	clr	ZH
	lsl	ZL ;calculate char address: ZL * 6 + charset base address
	rol	ZH
	movw	R1:R0, ZH:ZL
	lsl	ZL
	rol	ZH
	add	ZL, R0
	adc	ZH, R1
	adiw	ZH:ZL, charset*2
	ldi	temp, 6 ;6 pixels = 1 char
lcd_text_char_loop:
	dec	lcd_x
	lsl	byte
	lpm	R0, Z+
	and	R0, mask
	breq	PC+2
	ori	byte, 0x01
	mov	temp2, lcd_x ;send byte every 8 pixels
	andi	temp2, 0x07
	brne	PC+2
	rcall	lcd_data
	dec	temp
	brne	lcd_text_char_loop
	cpi	lcd_x, 0
	brne	lcd_text_line_loop
	
	subi	YL, 40
	inc	lcd_y
	mov	temp, lcd_y
	andi	temp, 0x07
	brne	lcd_text ;not done until LCD line is divisible by 8
	ret
	
;--------------------

lcd_clear:
	;clear LCD
	ldi	lcd_y, 0
	ldi	temp, 64
lcd_clear_lineloop:
	rcall	lcd_pos
	ldi	byte, 0x00
	ldi	temp2, 240/8
lcd_clear_byteloop:
	rcall	lcd_data
	dec	temp2
	brne	lcd_clear_byteloop
	inc	lcd_y
	cpi	lcd_y, 64
	brne	lcd_clear_lineloop
	ret

;--------------------
	
lcd_pos:
	;set LCD write address
	mov	byte, lcd_y ;set LCD Y address
	andi	byte, 0x0F
	ori	byte, 0x20
	rcall	lcd_cmd
	mov	byte, lcd_y
	swap	byte
	andi	byte, 0x0F
	ori	byte, 0x30
	rcall	lcd_cmd
	ldi	byte, 0x00 ;set LCD X address = 0
	rcall	lcd_cmd
	ldi	byte, 0x10
	rcall	lcd_cmd
	ret

;--------------------

lcd_cmd:
	;send command byte to LCD
	sbi	LCD_CTRL_PORT, LCD_RS
	rjmp	lcd_send
lcd_data:
	;send data byte to LCD
	cbi	LCD_CTRL_PORT, LCD_RS
lcd_send:
	out	LCD_DATA_PORT, byte
	sbi	LCD_CTRL_PORT, LCD_E
	cbi	LCD_CTRL_PORT, LCD_E
	ret

;--------------------

lcd_wait:
	;wait ~27ms
	clr	R0
	clr	R1
lcd_wait_loop:
	dec	R0
	brne	lcd_wait_loop
	dec	R1
	brne	lcd_wait_loop
	ret
	
